module Main where

import Data.Either (Either(..))
import Data.Foldable (for_)
import Data.Generic.Rep (class Generic)
import Data.Generic.Rep.Show (genericShow)
import Data.Maybe (Maybe(..))
import Data.Tuple (Tuple(..))
import Effect (Effect)
import Effect.Ref (Ref, new, read, write)
import Prelude (class Show, Unit, bind, discard, pure, show, unit, (+), (-), (<>), (>>=))
import Web.DOM.Document as D
import Web.DOM.Element as E
import Web.DOM.HTMLCollection as C
import Web.DOM.Node as N
import Web.Event.Event (Event, EventType(..))
import Web.Event.EventTarget (addEventListener, eventListener)
import Web.HTML.HTMLDocument (HTMLDocument, toDocument)
import Web.HTML (window)
import Web.HTML.Window as W

main :: Effect Unit
main = do
  w <- window
  d <- W.document w
  loadEvent <- eventListener (onLoadEvent d)
  addEventListener (EventType "load") loadEvent false (W.toEventTarget w)

onLoadEvent :: forall a. HTMLDocument -> a -> Effect Unit
onLoadEvent d e = do
  dialogs <- D.getElementsByClassName "fibonacci-dialog" (toDocument d) >>= C.toArray
  for_ dialogs setupDialog

setupDialog :: E.Element -> Effect Unit
setupDialog element = do
    maybe_output <- E.getElementsByClassName "output" element >>= C.item 0
    maybe_button <- E.getElementsByClassName "next" element >>= C.item 0
    case Tuple maybe_output maybe_button of
        Tuple (Just output) (Just button) -> do
            let setTexts a b = do
                                  N.setTextContent a (E.toNode output)
                                  N.setTextContent b (E.toNode button)
            setTexts "Program output goes here" "fibonacci 5"
            buttonRef <- new ((Tuple 5 Nothing) :: Tuple Int (Maybe (Either Int (Tuple Kont Int))))
            clickEvent <- eventListener (onButtonEvent buttonRef setTexts)
            addEventListener (EventType "click") clickEvent true (E.toEventTarget button)

            pure unit
        Tuple _ _ -> pure unit

onButtonEvent :: Ref (Tuple Int (Maybe (Either Int (Tuple Kont Int))))
              -> (String -> String -> Effect Unit) -> Event -> Effect Unit
onButtonEvent var update _ = do
    Tuple n state <- read var
    case state of
        Nothing -> do
            let res = fibonacci'''' n
            write (Tuple n (Just (Right res))) var
            update (show res) "next"
        Just (Right (Tuple k i)) -> do
            let res = apply' k i
            case res of
                Left _ -> do
                    let n' = n + 1
                    update (show res) ("fibonacci " <> show n')
                    write (Tuple n' (Just res)) var
                Right _ -> do
                    update (show res) "next"
                    write (Tuple n (Just res)) var
        Just (Left i) -> do
            let res = fibonacci'''' n
            write (Tuple n (Just (Right res))) var
            update (show res) "next"

fibonacci :: Int -> Int
fibonacci 0 = 0
fibonacci 1 = 1
fibonacci n = fibonacci (n-2) + fibonacci (n-1)

-- convert to continuation passing form
fibonacci' :: Int -> Int
fibonacci' n = fibonacciCont n (\v -> v)

fibonacciCont :: forall k. Int -> (Int -> k) -> k
fibonacciCont 0 k = k 0
fibonacciCont 1 k = k 1
fibonacciCont n k
    = fibonacciCont (n-2) (\v ->
        fibonacciCont (n-1) (\w ->
            k (v + w)))

fibonacci'' :: Int -> Int
fibonacci'' n = fibonacciCont' n fibonacciP0

fibonacciCont' :: forall k. Int -> (Int -> k) -> k
fibonacciCont' 0 k = k 0
fibonacciCont' 1 k = k 1
fibonacciCont' n k
    = fibonacciCont' (n-2) (fibonacciP1 {n, k})

fibonacciP0 :: forall k. k -> k
fibonacciP0 x = x

fibonacciP1 :: forall k. {n :: Int, k :: (Int -> k)} -> Int -> k
fibonacciP1 {n, k} v = fibonacciCont' (n-1)
                       (fibonacciP2 {v, k})

fibonacciP2 :: forall k. {v :: Int, k :: (Int -> k)} -> Int -> k
fibonacciP2 {v,k} w = k (v + w)

data Kont = Done
          | Next {n :: Int, k :: Kont}
          | Conc {v :: Int, k :: Kont}

fibonacci''' :: Int -> Int
fibonacci''' n = fibonacciCont'' n Done

fibonacciCont'' :: Int -> Kont -> Int
fibonacciCont'' 0 k = apply k 0
fibonacciCont'' 1 k = apply k 1
fibonacciCont'' n k = fibonacciCont'' (n-2) (Next {n,k})

apply :: Kont -> Int -> Int
apply Done s = s
apply (Next {n, k}) v = fibonacciCont'' (n-1) (Conc {v,k})
apply (Conc {v, k}) w = apply k (v + w)

derive instance genericKont :: Generic Kont _

instance showKont :: Show Kont where
    show x = genericShow x

fibonacci'''' :: Int -> (Tuple Kont Int)
fibonacci'''' n = fibonacciCont''' n Done

fibonacciCont''' :: Int -> Kont -> Tuple Kont Int
fibonacciCont''' 0 k = Tuple k 0
fibonacciCont''' 1 k = Tuple k 1
fibonacciCont''' n k = fibonacciCont''' (n-2) (Next {n,k})

apply' :: Kont -> Int -> Either Int (Tuple Kont Int)
apply' Done s = Left s
apply' (Next {n, k}) v = Right (fibonacciCont''' (n-1) (Conc {v,k}))
apply' (Conc {v, k}) w = Right (Tuple k (v + w))
